@@ -36,22 +36,22 @@ class TestView : public RemoteView
     Q_OBJECT
 
 public:
-    explicit TestView(QWidget *parent = 0, const QUrl &url = QUrl(), KConfigGroup configGroup = KConfigGroup());
+    explicit TestView(QWidget *parent = nullptr, const QUrl &url = QUrl(), KConfigGroup configGroup = KConfigGroup());
 
-    virtual ~TestView();
+    ~TestView() override;
 
-    virtual QSize framebufferSize();
-    QSize sizeHint() const;
+    QSize framebufferSize() override;
+    QSize sizeHint() const override;
 
-    virtual bool isQuitting();
-    virtual bool start();
-    HostPreferences* hostPreferences();
+    bool isQuitting() override;
+    bool start() override;
+    HostPreferences* hostPreferences() override;
 
 public Q_SLOTS:
-    virtual void switchFullscreen(bool on);
+    void switchFullscreen(bool on) override;
 
 protected:
-    bool eventFilter(QObject *obj, QEvent *event);
+    bool eventFilter(QObject *obj, QEvent *event) override;
 
 private:
     TestHostPreferences *m_hostPreferences;
@@ -65,11 +65,11 @@ class TestHostPreferences : public HostPreferences
 {
     Q_OBJECT
 public:
-    explicit TestHostPreferences(KConfigGroup configGroup, QObject *parent = 0)
+    explicit TestHostPreferences(KConfigGroup configGroup, QObject *parent = nullptr)
         : HostPreferences(configGroup, parent) {}
 
 protected:
-    virtual QWidget* createProtocolSpecificConfigPage() { return 0; };
+    QWidget* createProtocolSpecificConfigPage() override { return nullptr; };
 };
 
 #endif // TESTVIEW_H
