@@ -81,6 +81,7 @@ VncView::VncView(QWidget *parent, const QUrl &url, KConfigGroup configGroup)
     connect(&vncThread, SIGNAL(gotCut(QString)), this, SLOT(setCut(QString)), Qt::BlockingQueuedConnection);
     connect(&vncThread, SIGNAL(passwordRequest(bool)), this, SLOT(requestPassword(bool)), Qt::BlockingQueuedConnection);
     connect(&vncThread, SIGNAL(outputErrorMessage(QString)), this, SLOT(outputErrorMessage(QString)));
+    connect(&vncThread, &VncClientThread::gotCursor, this, [this](QCursor cursor){ setCursor(cursor); });
 
     m_clipboard = QApplication::clipboard();
     connect(m_clipboard, SIGNAL(dataChanged()), this, SLOT(clipboardDataChanged()));
@@ -115,7 +116,7 @@ bool VncView::eventFilter(QObject *obj, QEvent *event)
 
 QSize VncView::framebufferSize()
 {
-    return m_frame.size();
+    return m_frame.size() / devicePixelRatioF();
 }
 
 QSize VncView::sizeHint() const
@@ -134,8 +135,9 @@ void VncView::scaleResize(int w, int h)
 
     qCDebug(KRDC) << w << h;
     if (m_scale) {
-        m_verticalFactor = (qreal) h / m_frame.height();
-        m_horizontalFactor = (qreal) w / m_frame.width();
+        const QSize frameSize = m_frame.size() / m_frame.devicePixelRatio();
+        m_verticalFactor = (qreal) h / frameSize.height();
+        m_horizontalFactor = (qreal) w / frameSize.width();
 
 #ifndef QTONLY
         if (Settings::keepAspectRatio()) {
@@ -145,8 +147,8 @@ void VncView::scaleResize(int w, int h)
         m_verticalFactor = m_horizontalFactor = qMin(m_verticalFactor, m_horizontalFactor);
 #endif
 
-        const qreal newW = m_frame.width() * m_horizontalFactor;
-        const qreal newH = m_frame.height() * m_verticalFactor;
+        const qreal newW = frameSize.width() * m_horizontalFactor;
+        const qreal newH = frameSize.height() * m_verticalFactor;
         setMaximumSize(newW, newH); //This is a hack to force Qt to center the view in the scroll area
         resize(newW, newH);
     }
@@ -162,6 +164,11 @@ void VncView::updateConfiguration()
 
 void VncView::startQuitting()
 {
+    // Already quitted. No need to clean up again and also avoid triggering
+    // `disconnected` signal below.
+    if (m_quitFlag)
+        return;
+
     qCDebug(KRDC) << "about to quit";
 
     setStatus(Disconnecting);
@@ -197,6 +204,10 @@ void VncView::startQuitting()
 
     qCDebug(KRDC) << "Quit VNC thread success:" << quitSuccess;
 
+    // emit the disconnect siginal only after all the events are handled.
+    // Otherwise some error messages might be thrown away without
+    // showing to the user.
+    emit disconnected();
     setStatus(Disconnected);
 }
 
@@ -207,6 +218,11 @@ bool VncView::isQuitting()
 
 bool VncView::start()
 {
+    // This flag is used to make sure `startQuitting` only run once.
+    // This should not matter for now but it is an easy way to make sure
+    // things works in case the object may get reused.
+    m_quitFlag = false;
+
     QString vncHost = m_host;
     int vncPort = m_port;
 #ifdef LIBSSH_FOUND
@@ -243,10 +259,11 @@ bool VncView::start()
 #endif
 
     vncThread.setQuality(quality);
+    vncThread.setDevicePixelRatio(devicePixelRatioF());
 
     // set local cursor on by default because low quality mostly means slow internet connection
     if (quality == RemoteView::Low) {
-        showDotCursor(RemoteView::CursorOn);
+        showLocalCursor(RemoteView::CursorOn);
 #ifndef QTONLY
         // KRDC does always just have one main window, so at(0) is safe
         KXMLGUIClient *mainWindow = dynamic_cast<KXMLGUIClient*>(KMainWindow::memberList().at(0));
@@ -384,7 +401,7 @@ void VncView::outputErrorMessage(const QString &message)
     qCritical(KRDC) << message;
 
     if (message == QLatin1String("INTERNAL:APPLE_VNC_COMPATIBILTY")) {
-        setCursor(localDotCursor());
+        setCursor(localDefaultCursor());
         m_forceLocalCursor = true;
         return;
     }
@@ -428,7 +445,7 @@ void VncView::updateImage(int x, int y, int w, int h)
         setAttribute(Qt::WA_OpaquePaintEvent);
         installEventFilter(this);
 
-        setCursor(((m_dotCursorState == CursorOn) || m_forceLocalCursor) ? localDotCursor() : Qt::BlankCursor);
+        setCursor(((m_localCursorState == CursorOn) || m_forceLocalCursor) ? localDefaultCursor() : Qt::BlankCursor);
 
         setMouseTracking(true); // get mouse events even when there is no mousebutton pressed
         setFocusPolicy(Qt::WheelFocus);
@@ -438,8 +455,9 @@ void VncView::updateImage(int x, int y, int w, int h)
         if (m_scale) {
 #ifndef QTONLY
             qCDebug(KRDC) << "Setting initial size w:" <<m_hostPreferences->width() << " h:" << m_hostPreferences->height();
-            emit framebufferSizeChanged(m_hostPreferences->width(), m_hostPreferences->height());
-            scaleResize(m_hostPreferences->width(), m_hostPreferences->height());
+            QSize frameSize = QSize(m_hostPreferences->width(), m_hostPreferences->height()) / devicePixelRatioF();
+            emit framebufferSizeChanged(frameSize.width(), frameSize.height());
+            scaleResize(frameSize.width(), frameSize.height());
             qCDebug(KRDC) << "m_frame.size():" << m_frame.size() << "size()" << size();
 #else
 //TODO: qtonly alternative
@@ -460,7 +478,8 @@ void VncView::updateImage(int x, int y, int w, int h)
 #endif
     }
 
-    if ((y == 0 && x == 0) && (m_frame.size() != size())) {
+    const QSize frameSize = m_frame.size() / m_frame.devicePixelRatio();
+    if ((y == 0 && x == 0) && (frameSize != size())) {
         qCDebug(KRDC) << "Updating framebuffer size";
         if (m_scale) {
             setMaximumSize(QSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX));
@@ -468,14 +487,15 @@ void VncView::updateImage(int x, int y, int w, int h)
                 scaleResize(parentWidget()->width(), parentWidget()->height());
         } else {
             qCDebug(KRDC) << "Resizing: " << m_frame.width() << m_frame.height();
-            resize(m_frame.width(), m_frame.height());
-            setMaximumSize(m_frame.width(), m_frame.height()); //This is a hack to force Qt to center the view in the scroll area
-            setMinimumSize(m_frame.width(), m_frame.height());
-            emit framebufferSizeChanged(m_frame.width(), m_frame.height());
+            resize(frameSize);
+            setMaximumSize(frameSize); //This is a hack to force Qt to center the view in the scroll area
+            setMinimumSize(frameSize);
+            emit framebufferSizeChanged(frameSize.width(), frameSize.height());
         }
     }
 
-    repaint(QRectF(x * m_horizontalFactor, y * m_verticalFactor, w * m_horizontalFactor, h * m_verticalFactor).toAlignedRect());
+    const auto dpr = m_frame.devicePixelRatio();
+    repaint(QRectF(x / dpr * m_horizontalFactor, y / dpr * m_verticalFactor, w / dpr * m_horizontalFactor, h / dpr * m_verticalFactor).toAlignedRect());
 }
 
 void VncView::setViewOnly(bool viewOnly)
@@ -487,14 +507,22 @@ void VncView::setViewOnly(bool viewOnly)
     if (viewOnly)
         setCursor(Qt::ArrowCursor);
     else
-        setCursor(m_dotCursorState == CursorOn ? localDotCursor() : Qt::BlankCursor);
+        setCursor(m_localCursorState == CursorOn ? localDefaultCursor() : Qt::BlankCursor);
 }
 
-void VncView::showDotCursor(DotCursorState state)
+void VncView::showLocalCursor(LocalCursorState state)
 {
-    RemoteView::showDotCursor(state);
+    RemoteView::showLocalCursor(state);
 
-    setCursor(state == CursorOn ? localDotCursor() : Qt::BlankCursor);
+    if (state == CursorOn) {
+        // show local cursor, hide remote one
+        setCursor(localDefaultCursor());
+        vncThread.setShowLocalCursor(true);
+    } else {
+        // hide local cursor, show remote one
+        setCursor(Qt::BlankCursor);
+        vncThread.setShowLocalCursor(false);
+    }
 }
 
 void VncView::enableScaling(bool scale)
@@ -510,9 +538,10 @@ void VncView::enableScaling(bool scale)
         m_verticalFactor = 1.0;
         m_horizontalFactor = 1.0;
 
-        setMaximumSize(m_frame.width(), m_frame.height()); //This is a hack to force Qt to center the view in the scroll area
-        setMinimumSize(m_frame.width(), m_frame.height());
-        resize(m_frame.width(), m_frame.height());
+        const QSize frameSize = m_frame.size() / m_frame.devicePixelRatio();
+        setMaximumSize(frameSize); //This is a hack to force Qt to center the view in the scroll area
+        setMinimumSize(frameSize);
+        resize(frameSize);
     }
 }
 
@@ -537,9 +566,10 @@ void VncView::paintEvent(QPaintEvent *event)
     QPainter painter(this);
     painter.setRenderHint(QPainter::SmoothPixmapTransform);
 
+    const auto dpr = m_frame.devicePixelRatio();
     const QRectF dstRect = event->rect();
-    const QRectF srcRect(dstRect.x() / m_horizontalFactor, dstRect.y() / m_verticalFactor,
-                         dstRect.width() / m_horizontalFactor, dstRect.height() / m_verticalFactor);
+    const QRectF srcRect(dstRect.x() * dpr / m_horizontalFactor, dstRect.y() * dpr / m_verticalFactor,
+                         dstRect.width() * dpr / m_horizontalFactor, dstRect.height() * dpr / m_verticalFactor);
     painter.drawImage(dstRect, m_frame, srcRect);
 
     RemoteView::paintEvent(event);
@@ -599,7 +629,12 @@ void VncView::mouseEventHandler(QMouseEvent *e)
         }
     }
 
-    vncThread.mouseEvent(qRound(e->x() / m_horizontalFactor), qRound(e->y() / m_verticalFactor), m_buttonMask);
+    const auto dpr = devicePixelRatioF();
+    QPointF screenPos = e->screenPos();
+    // We need to restore mouse position in device coordinates.
+    // QMouseEvent::localPos() can be rounded (bug in Qt), but QMouseEvent::screenPos() is not.
+    QPointF pos = (e->pos() + (screenPos - screenPos.toPoint())) * dpr;
+    vncThread.mouseEvent(qRound(pos.x() / m_horizontalFactor), qRound(pos.y() / m_verticalFactor), m_buttonMask);
 }
 
 void VncView::wheelEventHandler(QWheelEvent *event)
